`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: 
// 
// Create Date: 12.02.2025 14:10:34
// Design Name: 
// Module Name: expo1
// Project Name: 
// Target Devices: 
// Tool Versions: 
// Description: 
// 
// Dependencies: 
// 
// Revision:
// Revision 0.01 - File Created
// Additional Comments:
// 
//////////////////////////////////////////////////////////////////////////////////
module expo_test #(parameter N=18, precision = 9) (clk,reset,sel,start,y);
input clk,reset,sel,start;
wire [N-1:0] x;
output [N-1:0] y;
wire tc,ena;
wire [1:0] ada;

ROM #(N,2) mem1(clk,ada,ena,x);
counterupN #(2) cnt(ada,'d0,tc,ena,clk,reset,tc,'d3);
pgbh pg(start,tc,ena,clk,reset);
/////////////// Reading.................

expo1 #(N,precision) uut(clk,reset,sel,x,y);
endmodule

module expo1 #(parameter N=18, precision = 9) (clk,reset,sel,x,y);
input clk,reset,sel;
input [N-1:0] x;
output [N-1:0] y;
wire [N-1:0] t [12:0];

scale3 #(N, precision) scale(clk,reset,x,t[0]);//   x/3

RSH_wired #(N,1) shift(x,t[12],);///  x/2


general_delay_block #(N,3) dly1(clk,reset,t[12],t[3]);
general_delay_block #(N,5) dly2(clk,reset,x,t[11]);

add_subN #(N) ad1(sel,'d512,t[0],t[1]);
regN #(N) rg1(t[2],clk,reset,t[1]);
multN #(N, precision) m1(t[2], t[3], t[4]);
regN #(N) rg2(t[5],clk,reset,t[4]);
add_subN #(N) ad2(sel,'d512,t[5],t[6]);//'d512 = 1
regN #(N) rg3(t[7],clk,reset,t[6]);
multN #(N, precision) m2(t[7], t[11], t[8]);
regN #(N) rg4(t[9],clk,reset,t[8]);
add_subN #(N) ad3(sel,'d512,t[9],t[10]);//'d512 = 1
regN #(N) rg5(y,clk,reset,t[10]);
endmodule

module scale3 #(parameter N=18, precision = 9) (clk,reset,x,y);
input clk,reset;
input [N-1:0] x;
output [N-1:0] y;

wire [N-1:0] t [8:0];

RSH_wired #(N,2) sh1(x,t[0],);
RSH_wired #(N,4) sh2(x,t[1],);
RSH_wired #(N,6) sh3(x,t[2],);
RSH_wired #(N,8) sh4(x,t[3],);

add_subN #(N) ad1(1'b0,t[0],t[1],t[4]);
add_subN #(N) ad2(1'b0,t[2],t[3],t[5]);

regN #(N) rg1(t[6],clk,reset,t[4]);
regN #(N) rg2(t[7],clk,reset,t[5]);

add_subN #(N) ad3(1'b0,t[6],t[7],t[8]);

regN #(N) rg3(y,clk,reset,t[8]);
endmodule

module add_subN #(parameter N=18)(sel,a,b,s);
input sel;
input [N-1:0] a,b;
output [N-1:0] s;///avoiding the carry out......

assign s = (sel)? (a-b):(a+b);
endmodule


module multN #(parameter N=18,//datawidth
				  parameter precision = 10)///accuracy
					(a, b, p);
  input signed [N-1:0] a,b;
  output [N -1:0] p;
  wire [2*N - 1:0] p1;
  assign p1 = a*b;///36 bits
  assign p = p1[N -1 + precision:precision];//18-bits
endmodule

module regN #(parameter N=18)(y,clk,reset,a);///PIPO...//N=1 means it is a DFF
    input [N-1:0] a;
    output reg [N-1:0] y;
    input clk,reset;
 
always @(posedge clk)
if (reset)///synchronous reset...//Active high
y <= 0;
else
y <= a;
endmodule

module RSH_wired #(parameter N = 18, p = 2)(a,b,rm);
    input [N-1:0] a;
    output [N-1:0] b;
    output [p-1:0] rm;
	 integer i;
	 reg [p:0] c;
assign rm = a[p-1:0];//1:0
assign b[N-1-p-1:0]= a[N-2:p];//18-1-2-1= 14
assign b[N-1:N-1-p] = c;
always @*
for (i= 0; i<=p; i = i+1) begin
c[i] = a[N-1];//msb
end
endmodule///

module general_delay_block #(parameter N=1,
									  parameter D = 4)(clk,reset,a,aN);
input clk,reset;
input [N-1:0] a;
output [N-1:0] aN;
wire [N-1:0] tmp [D:0];///D+1
assign tmp[0] = a;
generate
genvar p;
for (p = 1; p <= D; p = p+1) begin: General_delay
regN #(N) rg(tmp[p],clk,reset,tmp[p-1]);
end
endgenerate
assign aN = tmp[D];
endmodule


module ROM #(parameter N = 18,
							 parameter M = 2)
							(clk,ada,ena,outa);
////M is the address width
///N is the data width
input clk, ena;
input [M-1:0] ada;
//input [N - 1:0] ina;
output reg [N-1:0] outa;
reg [N-1:0] mem [0:2**M -1];

initial begin
 $readmemb("initial.mem", mem); 
end

always@(posedge clk)
if(ena) begin
    outa = mem[ada];
 end
 else
   outa = outa;
	
endmodule

module counterupN #(parameter N = 10)(count,data,load,en,clk,reset,tc,lmt);
output [N-1:0] count;
output reg tc;
input [N-1:0] data,lmt;
input load, en, clk,reset;
reg [N-1:0] count;

always @(posedge clk)
if (reset) begin
  count <= 1'b0 ;
end else if (load) begin
  count <= data;
end else if (en)
  count <= count + 1'b1;
else count <= count;
always @*
if (count ==lmt)
tc=1;
else tc=0;
endmodule

module pgbh(start,stop,q,clk,reset);///phase generation
	 input start,stop,clk,reset;
	 output reg q;
	 
always @(posedge clk)
begin
if (reset)
q = 0;
else if (start)
q = 1'b1;
else if(stop)
q = 1'b0;
end
endmodule